use super::*;

#[derive(Debug)]
struct BlueprintStateUnlockedInner {
    /// The global random number generator
    srng: StableRng,
    /// The unique id of this blueprint
    id: BlueprintStateId,
    /// The name of this blueprint state
    name: String,
}

#[derive(Debug)]
pub struct BlueprintStateIpv4Params {
    pub locations: NetworkLocationsList,
    pub prefix: WeightedList<u8>,
    pub gateway: Option<BlueprintStateGatewayParams>,
}

#[derive(Debug)]
pub struct BlueprintStateIpv6Params {
    pub locations: NetworkLocationsList,
    pub prefix: WeightedList<u8>,
    pub gateway: Option<BlueprintStateGatewayParams>,
}

#[derive(Debug)]
pub struct BlueprintStateGatewayParams {
    pub translation: WeightedList<config::Translation>,
    pub upnp: Probability,
    pub locations: Option<MachineLocationsList>,
}

#[derive(Debug)]
struct BlueprintStateIpv4 {
    params: BlueprintStateIpv4Params,
    gateway: Option<BlueprintStateIpv4Gateway>,
}

#[derive(Debug)]
struct BlueprintStateIpv6 {
    params: BlueprintStateIpv6Params,
    gateway: Option<BlueprintStateIpv6Gateway>,
}

#[derive(Debug)]
struct BlueprintStateIpv4Gateway {
    params: BlueprintStateGatewayParams,
}

#[derive(Debug)]
struct BlueprintStateIpv6Gateway {
    params: BlueprintStateGatewayParams,
}

#[derive(Debug)]
struct BlueprintStateInner {
    limit_network_count: Option<usize>,
    networks: Vec<NetworkStateId>,
    model: Option<WeightedList<String>>,
    ipv4: Option<BlueprintStateIpv4>,
    ipv6: Option<BlueprintStateIpv6>,
}

#[derive(Debug, Clone)]
pub struct BlueprintState {
    unlocked_inner: Arc<BlueprintStateUnlockedInner>,
    inner: Arc<Mutex<BlueprintStateInner>>,
}

pub type BlueprintStateId = StateId<BlueprintState>;

impl BlueprintState {
    pub fn new(
        srng: StableRng,
        id: BlueprintStateId,
        name: String,
    ) -> MachineRegistryResult<BlueprintState> {
        Ok(Self {
            unlocked_inner: Arc::new(BlueprintStateUnlockedInner { srng, id, name }),
            inner: Arc::new(Mutex::new(BlueprintStateInner {
                limit_network_count: None,
                networks: Vec::new(),
                model: None,
                ipv4: None,
                ipv6: None,
            })),
        })
    }

    pub fn set_limit_network_count(&self, limit_network_count: Option<usize>) {
        let mut inner = self.inner.lock();
        inner.limit_network_count = limit_network_count;
    }

    pub fn set_model(&self, model: WeightedList<String>) {
        let mut inner = self.inner.lock();
        inner.model = Some(model);
    }

    pub fn set_ipv4(
        &self,
        params: BlueprintStateIpv4Params,
        gateway_params: Option<BlueprintStateGatewayParams>,
    ) {
        let mut inner = self.inner.lock();
        if inner.ipv4.is_none() {
            inner.ipv4 = Some(BlueprintStateIpv4 {
                params,
                gateway: None,
            });
        } else {
            inner.ipv4.as_mut().map(|ipv4| {
                ipv4.params = params;
            });
        }
        let ipv4 = inner.ipv4.as_mut().expect("must exist");

        if ipv4.gateway.is_some() {
            if let Some(gateway_params) = gateway_params {
                ipv4.gateway.as_mut().expect("must exist").params = gateway_params;
            } else {
                ipv4.gateway = None;
            }
        } else if let Some(gateway_params) = gateway_params {
            ipv4.gateway = Some(BlueprintStateIpv4Gateway {
                params: gateway_params,
            })
        }
    }

    pub fn set_ipv6(
        &self,
        params: BlueprintStateIpv6Params,
        gateway_params: Option<BlueprintStateGatewayParams>,
    ) {
        let mut inner = self.inner.lock();
        if inner.ipv6.is_none() {
            inner.ipv6 = Some(BlueprintStateIpv6 {
                params,
                gateway: None,
            });
        } else {
            inner.ipv6.as_mut().map(|ipv6| {
                ipv6.params = params;
            });
        }
        let ipv6 = inner.ipv6.as_mut().expect("must exist");

        if ipv6.gateway.is_some() {
            if let Some(gateway_params) = gateway_params {
                ipv6.gateway.as_mut().expect("must exist").params = gateway_params;
            } else {
                ipv6.gateway = None;
            }
        } else if let Some(gateway_params) = gateway_params {
            ipv6.gateway = Some(BlueprintStateIpv6Gateway {
                params: gateway_params,
            })
        }
    }

    pub fn is_active(&self) -> MachineRegistryResult<bool> {
        let inner = self.inner.lock();

        // See if there's room for another network
        if let Some(limit_network_count) = inner.limit_network_count {
            if inner.networks.len() >= limit_network_count {
                return Ok(false);
            }
        }

        todo!("needs better implementation");

        Ok(true)
    }

    fn generate_model_inner(
        inner: &mut BlueprintStateInner,
        machine_registry_inner: &mut MachineRegistryInner,
        network_state: NetworkState,
    ) -> MachineRegistryResult<()> {
        let model_name = match inner.model.clone() {
            Some(models) => machine_registry_inner
                .srng()
                .weighted_choice_ref(&models)
                .clone(),
            None => machine_registry_inner.config().default_model.clone(),
        };
        let Some(model) = machine_registry_inner.config().models.get(&model_name) else {
            return Err(MachineRegistryError::ModelNotFound);
        };

        let params = NetworkStateModelParams {
            latency: Some(model.latency.clone()),
            distance: model.distance.clone(),
            loss: model.loss,
        };
        network_state.with_model(params);
        Ok(())
    }

    fn generate_ipv4_inner(
        inner: &mut BlueprintStateInner,
        machine_registry_inner: &mut MachineRegistryInner,
        network_state: NetworkState,
    ) -> MachineRegistryResult<()> {
        network_state.clear_ipv4(machine_registry_inner);
        let Some(ipv4) = inner.ipv4.as_ref() else {
            return Ok(());
        };

        // Get maximum prefix
        let max_prefix = ipv4
            .params
            .prefix
            .iter()
            .max()
            .copied()
            .expect("must have at least one element");

        // Get addresses for network
        let (subnet, super_net) = match &ipv4.params.locations {
            NetworkLocationsList::Allocations { allocations } => {
                // Get allocations which have subnets that would fit
                // our maximum requested prefix
                let Some(alloc_subnets) = allocations.try_filter_map(|allocation_name| {
                    let allocation = machine_registry_inner
                        .config()
                        .allocations
                        .get(allocation_name)
                        .expect("must exist");
                    Ok(allocation
                        .subnets
                        .subnet4
                        .as_ref()
                        .map(|subnet| subnet.filter(|p| p.prefix_len() <= max_prefix))
                        .flatten())
                })?
                else {
                    return Err(MachineRegistryError::NoAllocation);
                };

                // Pick an allocation
                let subnets = machine_registry_inner
                    .srng()
                    .weighted_choice_ref(&alloc_subnets);

                // Pick a subnet
                let net = *machine_registry_inner.srng().weighted_choice_ref(subnets);

                // Pick a prefix length that would fit in the subnet
                let opt_subnet = ipv4
                    .params
                    .prefix
                    .filter(|p| *p >= net.prefix_len())
                    .as_ref()
                    .map(|wl| {
                        let subnet_prefix =
                            machine_registry_inner.srng().weighted_choice_ref(wl).clone();

                        // Use an address pool temporarily to pick a subnet
                        let mut address_pool =
                            AddressPool::<()>::new();
                        address_pool.add_scope_v4(net);
                        address_pool.allocate_random_v4(machine_registry_inner.srng(), subnet_prefix, ())
                    })
                    .transpose()?
                    .flatten();
                let Some(subnet) = opt_subnet else {
                    return Err(MachineRegistryError::NoAllocation);
                };
                (subnet, None)
            }
            NetworkLocationsList::Networks { networks } => {
                // Get networks which have subnets that would fit
                // our maximum requested prefix
                let Some(available_networks) = networks.try_filter(|network_id| {
                    let network_state = machine_registry_inner
                        .network_states()
                        .get_state(*network_id)
                        .expect("must exist");

                    Ok(network_state.can_allocate_subnet_v4(None, max_prefix))
                })?
                else {
                    return Err(MachineRegistryError::NoAllocation);
                };

                // Pick a network
                let network_id = *machine_registry_inner
                    .srng()
                    .weighted_choice_ref(&available_networks);
                let network_state = machine_registry_inner
                    .network_states()
                    .get_state(network_id)
                    .expect("must exist");

                // Pick a prefix that fits in this network and allocate from it
                let opt_subnet = ipv4
                    .params
                    .prefix
                    .filter(|p| network_state.can_allocate_subnet_v4(None, *p))
                    .as_ref()
                    .map(|wl| {
                        let subnet_prefix =
                            machine_registry_inner.srng().weighted_choice_ref(wl).clone();

                        // Allocate subnet from this network
                        network_state.allocate_subnet_v4(
                            machine_registry_inner,
                            OwnerTag::Network(network_state.id()),
                            None,
                            subnet_prefix,
                        )
                    })
                    .transpose()?;
                let Some(subnet) = opt_subnet else {
                    return Err(MachineRegistryError::NoAllocation);
                };

                (subnet, Some(network_id))
            }
        };

        let params = NetworkStateIpv4Params {
            allocation: subnet,
            super_net,
        };

        let gateway_params = match ipv4.gateway.as_ref() {
            Some(v4gw) => {
                let translation = machine_registry_inner
                    .srng()
                    .weighted_choice_ref(&v4gw.params.translation)
                    .clone();
                let upnp = machine_registry_inner
                    .srng()
                    .probability_test(v4gw.params.upnp);

                let location = match v4gw.params.locations {
                    Some(locations) => todo!(),
                    None => todo!(),
                };

xxx instantiate and clean up on failure

                Some(NetworkStateIpv4GatewayParams {
                    translation,
                    upnp,
                    external_network: todo!(),
                    internal_address: None,
                    external_address: None,
                })
            }
            None => None,
        };

        network_state.set_ipv4(machine_registry_inner, params, gateway_params);
        Ok(())
    }

    fn generate_ipv6_inner(
        inner: &mut BlueprintStateInner,
        machine_registry_inner: &mut MachineRegistryInner,
        network_state: NetworkState,
    ) -> MachineRegistryResult<()> {
        //
    }

    pub fn generate(
        &self,
        machine_registry_inner: &mut MachineRegistryInner,
    ) -> MachineRegistryResult<NetworkState> {
        let mut inner = self.inner.lock();

        // See if there's room for another network
        if let Some(limit_network_count) = inner.limit_network_count {
            if inner.networks.len() >= limit_network_count {
                return Err(MachineRegistryError::BlueprintComplete);
            }
        }

        // Allocate a network id
        let network_state_id = machine_registry_inner.network_states_mut().allocate_id();

        // Create an anonymous network state
        let network_state = NetworkState::new(network_state_id, None);

        if let Err(e) = (|| {
            Self::generate_model_inner(&mut *inner, machine_registry_inner, network_state.clone())?;
            Self::generate_ipv4_inner(&mut *inner, machine_registry_inner, network_state.clone())?;
            Self::generate_ipv6_inner(&mut *inner, machine_registry_inner, network_state.clone())?;
            Ok(())
        })() {
            // Release the network state and id if things failed to allocate
            network_state.release(machine_registry_inner);
            machine_registry_inner
                .network_states_mut()
                .release_id(network_state_id);
            return Err(e);
        }

        // Attach the state to the id
        machine_registry_inner
            .network_states_mut()
            .attach_state(network_state.clone());

        // Record the newly instantiated network
        inner.networks.push(network_state_id);

        Ok(network_state)
    }

    pub fn for_each_network_id<F, R>(&self, callback: F) -> MachineRegistryResult<Option<R>>
    where
        F: Fn(NetworkStateId) -> MachineRegistryResult<Option<R>>,
    {
        let inner = self.inner.lock();
        for network_id in &inner.networks {
            if let Some(res) = callback(*network_id)? {
                return Ok(Some(res));
            }
        }
        Ok(None)
    }

    pub fn on_network_released(&self, network_id: NetworkStateId) {
        let mut inner = self.inner.lock();
        let pos = inner
            .networks
            .iter()
            .position(|id| *id == network_id)
            .expect("must exist");
        inner.networks.remove(pos);
    }
}

impl State for BlueprintState {
    fn id(&self) -> StateId<Self> {
        self.unlocked_inner.id.clone()
    }

    fn name(&self) -> Option<String> {
        Some(self.unlocked_inner.name.clone())
    }
}
