use super::*;

#[derive(Debug)]
struct TemplateStateImmutable {
    id: TemplateStateId,
    name: String,
}

#[derive(Debug, Clone)]
struct PerNetworkInfo {
    limit_machine_count: Option<usize>,
    machines: imbl::HashSet<MachineStateId>,
}

#[derive(Debug, Clone)]
enum BlueprintAvailability {
    Existing(NetworkState),
    Generate(BlueprintState),
}

#[derive(Debug, Clone)]
struct TemplateStateFields {
    limit_machine_count: Option<usize>,
    limit_machines_per_network: Option<WeightedList<usize>>,
    locations_list: Option<MachineLocationsList>,
    machines: imbl::HashSet<MachineStateId>,
    machines_per_network: imbl::HashMap<NetworkStateId, PerNetworkInfo>,
    disable_capabilities: imbl::Vector<Arc<String>>,
}

#[derive(Debug, Clone)]
pub struct TemplateState {
    immutable: Arc<TemplateStateImmutable>,
    fields: Arc<TemplateStateFields>,
}

pub type TemplateStateId = StateId<TemplateState>;

impl TemplateState {
    pub fn new(id: TemplateStateId, name: String) -> Self {
        Self {
            immutable: Arc::new(TemplateStateImmutable { id, name }),
            fields: Arc::new(TemplateStateFields {
                limit_machine_count: None,
                limit_machines_per_network: None,
                locations_list: None,
                machines: imbl::HashSet::new(),
                machines_per_network: imbl::HashMap::new(),
                disable_capabilities: imbl::Vector::new(),
            }),
        }
    }

    pub fn set_disable_capabilities(&mut self, disable_capabilities: Vec<String>) {
        let disable_capabilities =
            imbl::Vector::from_iter(disable_capabilities.into_iter().map(Arc::new));
        // Update fields
        self.fields = Arc::new(TemplateStateFields {
            disable_capabilities,
            ..(*self.fields).clone()
        });
    }

    pub fn set_networks_list(&mut self, networks: WeightedList<NetworkStateId>) {
        let locations_list = Some(MachineLocationsList::Networks { networks });

        // Update fields
        self.fields = Arc::new(TemplateStateFields {
            locations_list,
            ..(*self.fields).clone()
        });
    }

    pub fn set_blueprints_list(mut self, blueprints: WeightedList<BlueprintStateId>) {
        let locations_list = Some(MachineLocationsList::Blueprints { blueprints });

        // Update fields
        self.fields = Arc::new(TemplateStateFields {
            locations_list,
            ..(*self.fields).clone()
        });
    }

    pub fn clear_locations_list(&mut self) {
        let locations_list = None;

        // Update fields
        self.fields = Arc::new(TemplateStateFields {
            locations_list,
            ..(*self.fields).clone()
        });
    }

    pub fn set_limit_machine_count(&mut self, limit_machine_count: Option<usize>) {
        // Update fields
        self.fields = Arc::new(TemplateStateFields {
            limit_machine_count,
            ..(*self.fields).clone()
        });
    }

    pub fn set_limit_machines_per_network(
        &mut self,
        limit_machines_per_network: Option<WeightedList<usize>>,
    ) {
        // Update fields
        self.fields = Arc::new(TemplateStateFields {
            limit_machines_per_network,
            ..(*self.fields).clone()
        });
    }

    fn is_network_available(&self, network_state: NetworkState) -> MachineRegistryResult<bool> {
        // If the network is not active, it is not available
        if !network_state.is_active()? {
            return Ok(false);
        }

        // Get the per network info
        let Some(pni) = self.fields.machines_per_network.get(&network_state.id()) else {
            // If we haven't allocated anything in the network yet it is
            // by definition available
            return Ok(true);
        };

        // If this template has allocated the maximum number of machines per-network
        // for this network, then it is not available
        if let Some(limit_machine_count) = pni.limit_machine_count {
            if pni.machines.len() >= limit_machine_count {
                return Ok(false);
            }
        }

        Ok(true)
    }

    fn is_blueprint_available(
        &self,
        machine_registry_inner: &MachineRegistryInner,
        blueprint_state: BlueprintState,
    ) -> MachineRegistryResult<Option<BlueprintAvailability>> {
        // See if the networks generated from this blueprint so far have availability
        // in this template
        if let Some(available_network_state) = blueprint_state.for_each_network_id(|id| {
            // Check the network's availability
            let network_state = machine_registry_inner.network_states().get_state(id)?;
            if self.is_network_available(network_state.clone())? {
                // We found one
                return Ok(Some(network_state));
            }
            // Try next network
            Ok(None)
        })? {
            // We found a usable network
            return Ok(Some(BlueprintAvailability::Existing(
                available_network_state,
            )));
        }

        // If the blueprint is active, it is available because it can make a new network
        if blueprint_state.is_active()? {
            return Ok(Some(BlueprintAvailability::Generate(blueprint_state)));
        }

        Ok(None)
    }

    pub fn is_active(
        &self,
        machine_registry_inner: &MachineRegistryInner,
    ) -> MachineRegistryResult<bool> {
        // See if we have reached our machine limit
        if let Some(limit_machine_count) = self.fields.limit_machine_count {
            if self.fields.machines.len() >= limit_machine_count {
                return Ok(false);
            }
        }

        let Some(locations_list) = self.fields.locations_list.as_ref() else {
            return Ok(false);
        };

        match locations_list {
            MachineLocationsList::Networks { networks } => {
                // Filter the weighted list of networks to those that are still active and or not yet started
                if networks
                    .try_filter(|id| {
                        let network_state =
                            machine_registry_inner.network_states().get_state(*id)?;
                        self.is_network_available(network_state)
                    })?
                    .is_none()
                {
                    return Ok(false);
                };
            }
            MachineLocationsList::Blueprints { blueprints } => {
                // Filter the weighted list of blueprints to those that are still active or not yet started and can allocate
                if blueprints
                    .try_filter(|id| {
                        let blueprint_state =
                            machine_registry_inner.blueprint_states().get_state(*id)?;

                        self.is_blueprint_available(machine_registry_inner, blueprint_state)
                            .map(|x| x.is_some())
                    })?
                    .is_none()
                {
                    return Ok(false);
                };
            }
        };

        Ok(true)
    }

    pub fn generate(
        &mut self,
        machine_registry_inner: &mut MachineRegistryInner,
    ) -> MachineRegistryResult<MachineStateId> {
        // See if we have reached our machine limit
        if let Some(limit_machine_count) = self.fields.limit_machine_count {
            if self.fields.machines.len() < limit_machine_count.try_into().unwrap_or(usize::MAX) {
                return Err(MachineRegistryError::TemplateComplete);
            }
        }

        // If existing networks are all full, we'd have to allocate one, see if we'd be able to do that
        let Some(locations_list) = self.fields.locations_list.as_ref() else {
            return Err(MachineRegistryError::TemplateComplete);
        };

        // Get a network to generate the machine on
        let network_state = match locations_list {
            MachineLocationsList::Networks { networks } => {
                // Filter the weighted list of networks to those that are still active and or not yet started
                let Some(available_networks) = networks.try_filter_map(|id| {
                    let network_state = machine_registry_inner.network_states().get_state(*id)?;
                    if self.is_network_available(network_state.clone())? {
                        Ok(Some(network_state))
                    } else {
                        Ok(None)
                    }
                })?
                else {
                    return Err(MachineRegistryError::NetworkComplete);
                };

                // Weighted choice of network now that we have a candidate list
                let network_state = machine_registry_inner
                    .srng()
                    .weighted_choice_ref(&available_networks);

                // Return network state to use
                network_state.clone()
            }
            MachineLocationsList::Blueprints { blueprints } => {
                // Filter the weighted list of blueprints to those that are still active or not yet started and can allocate
                let Some(available_blueprints) = blueprints.try_filter_map(|id| {
                    let blueprint_state =
                        machine_registry_inner.blueprint_states().get_state(*id)?;

                    self.is_blueprint_available(machine_registry_inner, blueprint_state)
                })?
                else {
                    return Err(MachineRegistryError::BlueprintComplete);
                };

                // Weighted choice of blueprint now that we have a candidate list
                match machine_registry_inner
                    .srng()
                    .weighted_choice_ref(&available_blueprints)
                {
                    BlueprintAvailability::Existing(network_state) => network_state.clone(),
                    BlueprintAvailability::Generate(blueprint_state) => {
                        blueprint_state.generate(machine_registry_inner)?
                    }
                }
            }
        };

        // Allocate a machine id
        let machine_state_id = machine_registry_inner.machine_states_mut().allocate_id();

xxx continue here

        // Create an anonymous machine state
        let machine_state = MachineState::new(machine_state_id, None);

        if let Err(e) = (|| {
            // Build out the machine state from the template
            machine_state.set_disable_capabilities(inner.disable_capabilities.clone());
            machine_state.set_bootstrap(false);

            // Make the default route interface
            let vin0 = machine_state.allocate_interface(None, None)?;
            machine_state.attach_network(machine_registry_inner, &vin0, network_state.id())?;
            if network_state.is_ipv4() {
                machine_state.allocate_address_ipv4(machine_registry_inner, &vin0, None, None)?;
            }
            if network_state.is_ipv6() {
                machine_state.allocate_address_ipv6(machine_registry_inner, &vin0, None, None)?;
            }
            Ok(())
        })() {
            // Release the machine state and id if things failed to allocate
            machine_state.release(machine_registry_inner);
            machine_registry_inner
                .machine_states_mut()
                .release_id(machine_state_id);
            return Err(e);
        }

        // Attach the state to the id
        machine_registry_inner
            .machine_states_mut()
            .attach_state(machine_state.clone());

        // Record the newly instantiated machine
        inner.machines.insert(machine_state_id);
        let limit_machines_per_network = inner.limit_machines_per_network.clone();
        let per_network_info = inner
            .machines_per_network
            .entry(network_state.id())
            .or_insert_with(|| {
                let limit_machine_count = limit_machines_per_network.map(|wl| {
                    machine_registry_inner
                        .srng()
                        .weighted_choice_ref(&wl)
                        .clone()
                });
                PerNetworkInfo {
                    limit_machine_count,
                    machines: HashSet::new(),
                }
            });
        per_network_info.machines.insert(machine_state_id);

        Ok(machine_state)
    }

    pub fn on_machine_released(&self, machine_state_id: MachineStateId) {
        let mut inner = self.inner.lock();
        inner.machines.remove(&machine_state_id);
        for (_network_id, pni) in &mut inner.machines_per_network {
            pni.machines.remove(&machine_state_id);
        }
    }
}

impl State for TemplateState {
    fn id(&self) -> StateId<Self> {
        self.unlocked_inner.id.clone()
    }

    fn name(&self) -> Option<String> {
        Some(self.unlocked_inner.name.clone())
    }
}
